#!/usr/bin/env python
"""
__WebTools__

Core of the DMWM web stack. Provides a basic set of features including security,
performance optimisations (code minification, zipping and correct headers and
etags)
"""
