#!/usr/bin/env python
"""
_GetOutputParentLFNs_

Oracle implementation of Jobs.GetOutputParentLFNs
"""

from WMCore.WMBS.MySQL.Jobs.GetOutputParentLFNs import GetOutputParentLFNs as GetOutputParentLFNsMySQL

class GetOutputParentLFNs(GetOutputParentLFNsMySQL):
    pass
