#!/usr/bin/env python
"""
_SetOutcomeBulk_

Oracle implementation of Jobs.SetOutcomeBulk
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.SetOutcomeBulk import SetOutcomeBulk as MySQLSetOutcomeBulk

class SetOutcomeBulk(MySQLSetOutcomeBulk):
    pass
