@CHARSET "UTF-8";
body{ 
	margin: 0.63em; 
	padding: 0.63em; 
	padding-top:3em;
	background:rgb(255,255,255);
	clear:both;
}
.right{
	float: right;
	position: relative;
	padding-right: 1.25em;
}
.left{
	float: left;
	position: relative;
	padding-left: 1.25em;
}
p{
	font-size: 0.75em; 
	font-family: 'Lucida Grande', verdana, helvetica, arial, sans-serif;
	color: rgb(0, 0, 0);
	margin: 0.63em; 
	padding:0px; 
}
h1, h2, h3, h4, h5, h6{
	color: rgb(0, 0, 0);
	font-family: 'Lucida Grande', verdana, helvetica, arial, sans-serif; 
	font-weight:bold; 
	margin: 0.63em; 
	padding:0px; 
}
h1, h3, h5{
	text-decoration: underline;
}
h1, h2{	
	font-size: 2em; 
}
h3, h4{
	font-size: 1.5em;
}
h5, h6{
	font-size: 1em; 
}
.inline{
	display: inline;
}
.wide{
	width:100%;
}
.center{
	text-align: center
}