"""
_WorkUnit.ExistsByID_

Oracle implementation of WorkUnit.ExistsByID
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.ExistsByID import ExistsByID as ExistsByIDMySQL


class ExistsByID(ExistsByIDMySQL):
    """
    _WorkUnit.ExistsByID_

    Oracle implementation of WorkUnit.ExistsByID
    """

    pass
